c----------------------------------------------------------------------c
c                                                                      c
c                        Programma   T P R S                           c
c                                                                      c
c                 Teaching Purpose Reactor Simulation                  c
c                                                                      c
c                     W. Ambrosini, Ottobre 1997                       c
c                                                                      c
c----------------------------------------------------------------------c
      program core
      implicit double precision (a-h,o-z)
      character*80 riga
c
      parameter (m = 101)
      common /generl/ h,aleng,rcore,rrefl,omega,n,method,istvid
      common /nuclea/ diff(m,m),sigma(m,m),anisif(m,m),source(m,m)
      common /flux/ phi(0:m,0:m),oldphi(0:m,0:m),akeff
      common /winds/ ce(m,m),cw(m,m),cn(m,m),cs(m,m),co(m,m)
      dimension oldsou(m,m),class(m,m)
c
      open (unit=5,file='core.dat')
      open (unit=6,file='core.out')
      open (unit=7,file='core.pla')
      open (unit=8,file='core.txt')
c
c   lettura dei dati generali
c
      read(5,100) riga
      read(5,*) aleng,rcore,rrefl,n,method,omega
c
c   primo processamento per ottenere costanti di interesse
c   (first pre-processing for preparing relevant constants)
c
      h = aleng / dfloat (n)
      halfal = 0.5d00 * aleng
      halfh = 0.5d00 * h
      pi = 4.d00 * datan (1.d00)
c
c   lettura delle costanti nucleari del nocciolo
c   (reading the nuclear constants of the reactor core)
c
      read(5,100) riga
      read(5,*) dcore,sigcor,anisic
c
c   lettura delle costanti nucleari del riflettore
c   (option for printing on the video)
c
      read(5,100)
      read(5,*) drefl,sigref
c
c   opzione di stampa a video
c   (option for printing on the video)
c
      write(*,*) ' Video printing ? (1=yes; 0=no) '
      read(*,*) istvid
c
c   assegnazione delle costanti nucleari nei nodi
c   (assigning nuclear constants in the nodes)
c
      sumsou = 0.d00
      do 10 i = 1,n
      xi = h * dfloat(i) - halfh
      deltxi = xi - halfal
c
      do 10 j = 1,n
      yj = h * dfloat(j) - halfh
      deltyj = yj - halfal
c
      radius = dsqrt ( deltxi * deltxi + deltyj * deltyj )
c
c  nodi del nocciolo
c  (core nodes)
         if(radius.le.rcore) then
         diff(i,j) = dcore
         sigma(i,j) = sigcor
         anisif(i,j) = anisic
         ratr = radius / rcore
c
         soucor = 1.84d13
         source(i,j) = soucor * dcos ( 0.5d00 * pi * ratr )
         sumsou = sumsou + source(i,j)
         oldsou(i,j) = source(i,j)
         class(i,j) = 2.d00
c
c  nodi del riflettore
c  (reflector nodes)
         elseif(radius.le.rrefl) then
         diff(i,j) = drefl
         sigma(i,j) = sigref
         source(i,j) = 0.d00
         class(i,j) = 1.d00
c
c  nodi esterni
c  (external nodes)
         else
         diff(i,j) = 1.d05
         sigma(i,j) = 1.d05
         source(i,j) = 0.d00
         class(i,j) = 0.d00
         endif
c
   10 continue
c
c   assegnazione delle costanti in nodi particolari (npart > 0)
c   (assigning constants in particular nodes (if npart > 0)
c
      read(5,100) riga
      read(5,*) npart
c
         if(npart.gt.0) then
         read(5,100) riga
         do 20 ipart = 1,npart
         read(5,*) ip,jp,diff(ip,jp),sigma(ip,jp),anisif(ip,jp)
         xip = h * dfloat(ip) - halfh
         yjp = h * dfloat(jp) - halfh
         class(ip,jp) = 3.d00
   20    continue
         endif
c
c   costruzione dei coefficienti dell'equazione della diffusione
c   (the coefficients of the diffusion equation are set up)
c
      do 30 i = 1,n
      do 30 j = 1,n
c
         if(i.eq.1) then
         cw(i,j) = 2.d00 * diff(i,j)
         else
         cw(i,j) = 2.d00 * diff(i,j) * diff(i-1,j)
     &                   / ( diff(i,j) + diff(i-1,j) )
         endif
c
         if(i.eq.n) then
         ce(i,j) = 2.d00 * diff(i,j)
         else
         ce(i,j) = 2.d00 * diff(i,j) * diff(i+1,j)
     &                   / ( diff(i,j) + diff(i+1,j) )
         endif
c
         if(j.eq.1) then
         cs(i,j) = 2.d00 * diff(i,j)
         else
         cs(i,j) = 2.d00 * diff(i,j) * diff(i,j-1)
     &                   / ( diff(i,j) + diff(i,j-1) )
         endif
c
         if(j.eq.n) then
         cn(i,j) = 2.d00 * diff(i,j)
         else
         cn(i,j) = 2.d00 * diff(i,j) * diff(i,j+1)
     &                   / ( diff(i,j) + diff(i,j+1) )
         endif
c
      co(i,j) = ce(i,j) + cw(i,j) + cn(i,j) + cs(i,j)
     &                  + h * h * sigma(i,j)
c
   30 continue
c
c   iterazioni esterne
c   (external iterations)
c
      call gettim(ihr,imin,isec,icent)
      time0 = 3600.d00 * ihr + 60.d00 * imin + isec + 0.01d00 * icent
c
      akeff = 1.d00
      do 50 itext = 1,100000
c
c   soluzione del sistema con il metodo prescelto
c   (algebraic system solution with the selected method)
c
      if(method.eq.1) call jacobi (iter,rspect,itext)
      if(method.eq.2) call gseid (iter,itext)
      if(method.eq.3) call sor (iter,itext)
      if(method.eq.4) call lor (iter,itext)
c
      oldak = akeff
      oldsum = sumsou
c
      sumsou = 0.d00
      do 40 i = 1,n
      do 40 j = 1,n
      source(i,j) = anisif(i,j) * phi(i,j)
      sumsou = sumsou + source(i,j)
   40 continue
c
      akeff = oldak * sumsou / oldsum
c
         if(istvid.eq.1) then
         write(*,120) itext,akeff
         endif
      write(6,120) itext,akeff
c
      if(dabs(akeff-oldak).lt.1.d-7) goto 55
   50 continue
c
c
   55 continue
c
      call gettim(ihr,imin,isec,icent)
      time1 = 3600.d00 * ihr + 60.d00 * imin + isec + 0.01d00 * icent
      tcpu = time1 - time0
c
          if(method.eq.1) then
          omeopt = 2.d00 / ( 1.d00 + dsqrt ( 1.d00 - rspect * rspect ) )
               if(istvid.eq.1) then
               write(*,130) rspect
               write(*,135) omeopt
               endif
          write(6,130) rspect
          write(6,135) omeopt
          endif
c
c   scrittura dei valori del flusso
c
      area = 0.d00
      totpow = 0.d00
      powmax = 0.d00
      phimed = 0.d00
      sumasf = 0.d00
      do 60 i = 1,n
      xi = h * dfloat(i) - halfh
      do 60 j = 1,n
      yj = h * dfloat(j) - halfh
c
      sumasf = sumasf + anisif(i,j)
      phiasf = phi(i,j) * anisif(i,j)
      pow = phiasf / ( 2.5d00 * 3.1e10 )
      phimed = phimed + phiasf
      totpow = totpow + pow * h * h * 365.76d00
      if(powmax.lt.pow) powmax = pow
c
      write(7,160) xi,yj,class(i,j)
   60 write(8,110) xi,yj,phi(i,j)
      phimed = phimed / sumasf
      powavg = totpow / ( pi * rcore * rcore * 365.76d00 )
      fattpc = powmax / powavg
      react = ( akeff - 1.d00 ) / akeff
c
         if(istvid.eq.1) then
         write(*,170) react
         write(*,180) totpow
         write(*,190) phimed
         write(*,140) fattpc
         write(*,150) tcpu
         endif
c
      write(6,121) itext,akeff
      write(6,170) react
      write(6,180) totpow
      write(6,190) phimed
      write(6,140) fattpc
      write(6,150) tcpu
c
      stop
  100 format (a80)
  110 format (3(1x,e14.7))
  120 format (/,1x,' External Iteration n. ',i5,' Keff = ',f12.9)
  121 format (/,1x,' External Iteration number = ',i5,' Keff = ',f12.9)
  130 format (/,1x,' Jacobi matrix spectral radius = ',f12.9)
  135 format (1x,' Optimal Overrelaxation Parameter = ',f12.9)
  140 format (1x,' Radial Peaking Factor = ',f12.9)
  150 format (/,1x,' Processing time = ',f7.2,' s ')
  160 format (3(1x,e14.7))
  170 format (/,1x,' Reactivity = ',f12.9)
  180 format (1x,' Total Thermal Power = ',1pe14.7,' W ')
  190 format (1x,' Average Flux = ',1pe14.7,' n/(cm2.s)')
      end
c----------------------------------------------------------------------c
c                                                                      c
c     Soluzione delle Equazioni della Diffusione con Jacobi            c
c     (Solving the diffusion equations with the Jacobi method)         c
c                                                                      c
c----------------------------------------------------------------------c
      subroutine jacobi (iter,rspect,itext)
      implicit double precision (a-h,o-z)
c
      parameter (m = 101)
      common /generl/ h,aleng,rcore,rrefl,omega,n,method,istvid
      common /nuclea/ diff(m,m),sigma(m,m),anisif(m,m),source(m,m)
      common /flux/ phi(0:m,0:m),oldphi(0:m,0:m),akeff
      common /winds/ ce(m,m),cw(m,m),cn(m,m),cs(m,m),co(m,m)
c
c   assegnazione dell'approssimazione iniziale
c   (the initial approximation is assigned)
c
      np1 = n + 1
      h2 = h * h
      do 10 i = 0,np1
      do 10 j = 0,np1
c
       if(itext.eq.1) then
         if( (i.eq.0).or.(j.eq.0).or.(i.eq.np1).or.(j.eq.np1) ) then
         phi(i,j) = 0.d00
         oldphi(i,j) = 0.d00
         else
         phi(i,j) = h2 * source(i,j) / co(i,j) / akeff
         oldphi(i,j) = phi(i,j)
         endif
       else
       oldphi(i,j) = phi(i,j)
       endif
c
   10 continue
c
      oldel2 = 1.d00
      do 30 iter = 1,10000000
c
      ratmax = 0.d00
      del2su = 0.d00
c
      do 20 j = 1,n
      do 20 i = 1,n
      tm = ce(i,j) * oldphi(i+1,j) + cn(i,j) * oldphi(i,j+1)
     &   + cw(i,j) * oldphi(i-1,j) + cs(i,j) * oldphi(i,j-1)
c
      phi(i,j) = ( tm + h2 * source(i,j) / akeff ) / co(i,j)
c
      delphi = phi(i,j) - oldphi(i,j)
      del2su = del2su + delphi * delphi
c     ratio = dabs (delphi) / ( phi(i,j) + 1.d-10 )
      ratio = dabs (delphi)
      if(ratio.gt.ratmax) ratmax = ratio
c
   20 continue
c
      do 25 i = 1,n
      do 25 j = 1,n
      oldphi(i,j) = phi(i,j)
   25 continue
c
      rspect = dsqrt ( del2su / oldel2 )
      oldel2 = del2su
c
      if((ratmax.lt.1.d00).and.(iter.ne.1)) goto 35
c
      if(istvid.eq.1) write(*,101) iter,ratmax
      write(6,101) iter,ratmax
   30 continue
      write(6,100)
c
   35 return
  100 format (/,1x,'<<<<< Warning: Convergence Problems >>>>>')
  101 format (' Jacobi Internal Iteration n. ',i6,
     & ' Max. Error = ',e14.7)
      end
c----------------------------------------------------------------------c
c                                                                      c
c     Soluzione delle Equazioni della Diffusione con Gauss-Seidel      c
c     (Solving the diffusion equations with the Gauss-Seidel method)   c
c                                                                      c
c----------------------------------------------------------------------c
      subroutine gseid (iter,itext)
      implicit double precision (a-h,o-z)
c
      parameter (m = 101)
      common /generl/ h,aleng,rcore,rrefl,omega,n,method,istvid
      common /nuclea/ diff(m,m),sigma(m,m),anisif(m,m),source(m,m)
      common /flux/ phi(0:m,0:m),oldphi(0:m,0:m),akeff
      common /winds/ ce(m,m),cw(m,m),cn(m,m),cs(m,m),co(m,m)
c
c   assegnazione dell'approssimazione iniziale
c   (the initial approximation is assigned)
c
      np1 = n + 1
      h2 = h * h
      do 10 i = 0,np1
      do 10 j = 0,np1
c
       if(itext.eq.1) then
         if( (i.eq.0).or.(j.eq.0).or.(i.eq.np1).or.(j.eq.np1) ) then
         phi(i,j) = 0.d00
         oldphi(i,j) = 0.d00
         else
         phi(i,j) = h2 * source(i,j) / co(i,j) / akeff
         oldphi(i,j) = phi(i,j)
         endif
       else
       oldphi(i,j) = phi(i,j)
       endif
c
   10 continue
c
      do 30 iter = 1,10000000
c
      ratmax = 0.d00
c
      do 20 j = 1,n
      do 20 i = 1,n
      tm = ce(i,j) * phi(i+1,j) + cn(i,j) * phi(i,j+1)
      tmp1 = cw(i,j) * phi(i-1,j) + cs(i,j) * phi(i,j-1)
c
      phi(i,j) = ( tm + tmp1 + h2 * source(i,j) / akeff ) / co(i,j)
c
      delphi = phi(i,j) - oldphi(i,j)
c     ratio = dabs (delphi) / ( phi(i,j) + 1.d-10 )
      ratio = dabs (delphi)
      if(ratio.gt.ratmax) ratmax = ratio
c
      oldphi(i,j) = phi(i,j)
c
   20 continue
c
      if((ratmax.lt.1.d00).and.(iter.ne.1)) goto 35
c
      if(istvid.eq.1) write(*,101) iter,ratmax
      write(6,101) iter,ratmax
   30 continue
      write(6,100)
c
   35 return
  100 format (/,1x,'<<<<< Warning: Convergence Problems >>>>>')
  101 format (' GS Internal Iteration n. ',i6,
     & ' Max. Error = ',e14.7)
      end
c----------------------------------------------------------------------c
c                                                                      c
c     Soluzione delle Equazioni della Diffusione con SOR               c
c     (Solving the diffusion equations with the SOR method)            c
c                                                                      c
c----------------------------------------------------------------------c
      subroutine sor (iter,itext)
      implicit double precision (a-h,o-z)
c
      parameter (m = 101)
      common /generl/ h,aleng,rcore,rrefl,omega,n,method,istvid
      common /nuclea/ diff(m,m),sigma(m,m),anisif(m,m),source(m,m)
      common /flux/ phi(0:m,0:m),oldphi(0:m,0:m),akeff
      common /winds/ ce(m,m),cw(m,m),cn(m,m),cs(m,m),co(m,m)
c
c   assegnazione dell'approssimazione iniziale
c   (the initial approximation is assigned)
c
      np1 = n + 1
      h2 = h * h
      do 10 i = 0,np1
      do 10 j = 0,np1
c
       if(itext.eq.1) then
         if( (i.eq.0).or.(j.eq.0).or.(i.eq.np1).or.(j.eq.np1) ) then
         phi(i,j) = 0.d00
         oldphi(i,j) = 0.d00
         else
         phi(i,j) = h2 * source(i,j) / co(i,j) / akeff
         oldphi(i,j) = phi(i,j)
         endif
       else
       oldphi(i,j) = phi(i,j)
       endif
c
   10 continue
c
      do 30 iter = 1,10000000
c
      ratmax = 0.d00
c
      do 20 j = 1,n
      do 20 i = 1,n
      tm = ce(i,j) * phi(i+1,j) + cn(i,j) * phi(i,j+1)
      tmp1 = cw(i,j) * phi(i-1,j) + cs(i,j) * phi(i,j-1)
c
      phi(i,j) = ( tm + tmp1 + h2 * source(i,j) / akeff ) / co(i,j)
c
c   sovrarilassamento
c   (overrelaxation)
c
      phi(i,j) = ( phi(i,j) - oldphi(i,j) ) * omega + oldphi(i,j)
c
      delphi = phi(i,j) - oldphi(i,j)
c     ratio = dabs (delphi) / ( phi(i,j) + 1.d-10 )
      ratio = dabs (delphi)
      if(ratio.gt.ratmax) ratmax = ratio
c
      oldphi(i,j) = phi(i,j)
c
   20 continue
c
      if((ratmax.lt.1.d00).and.(iter.ne.1)) goto 35
c
      if(istvid.eq.1) write(*,101) iter,ratmax
      write(6,101) iter,ratmax
   30 continue
      write(6,100)
c
   35 return
  100 format (/,1x,'<<<<< Warning: Convergence Problems >>>>>')
  101 format (' SOR Internal Iteration n. ',i6,
     & ' Max. Error = ',e14.7)
      end
c----------------------------------------------------------------------c
c                                                                      c
c     Soluzione delle Equazioni della Diffusione con LOR               c
c     (Solving the diffusion equations with the LOR method)            c
c                                                                      c
c----------------------------------------------------------------------c
      subroutine lor (iter,itext)
      implicit double precision (a-h,o-z)
c
      parameter (m = 101)
      common /generl/ h,aleng,rcore,rrefl,omega,n,method,istvid
      common /nuclea/ diff(m,m),sigma(m,m),anisif(m,m),source(m,m)
      common /flux/ phi(0:m,0:m),oldphi(0:m,0:m),akeff
      common /winds/ ce(m,m),cw(m,m),cn(m,m),cs(m,m),co(m,m)
c
      dimension a(m),b(m),c(m),d(m),v(m),alef(m),bet(m)
c
c   assegnazione dell'approssimazione iniziale
c   (the initial approximation is assigned)
c
      np1 = n + 1
      h2 = h * h
      do 10 i = 0,np1
      do 10 j = 0,np1
c
       if(itext.eq.1) then
         if( (i.eq.0).or.(j.eq.0).or.(i.eq.np1).or.(j.eq.np1) ) then
         phi(i,j) = 0.d00
         oldphi(i,j) = 0.d00
         else
         phi(i,j) = h2 * source(i,j) / co(i,j) / akeff
         oldphi(i,j) = phi(i,j)
         endif
       else
       oldphi(i,j) = phi(i,j)
       endif
c
   10 continue
c
      do 40 iter = 1,10000000
c
      ratmax = 0.d00
c
c  ciclo sull'indice che definisce le linee orizzontali
c  (loop on the index defining the horizontal lines)
c
      do 30 j = 1,n
c
c  assegnazione dei coefficienti della matrice tridiagonale di linea
c  (assigning the coefficients of the tridiagonal matrix for the line)
c
      do 20 i = 1,n
      a(i) = - cw(i,j)
      b(i) = co(i,j)
      c(i) = - ce(i,j)
      d(i) = omega * ( h2 * source(i,j) / akeff
     &                 + cs(i,j) * phi(i,j-1) + cn(i,j) * phi(i,j+1) )
     &  + (1.d00 - omega) * ( - cw(i,j) * phi(i-1,j)
     &            + co(i,j) * phi(i,j) - ce(i,j) * phi(i+1,j) )
   20 continue
c
c  soluzione del sistema tridiagonale
c  (solution of the tri-diagonal system)
c
      call tdma (a,b,c,d,v,alef,bet,n,m)
c
c  assegnazione dei valori del flusso sulla linea j = costante
c  e valutazione dell'errore massimo
c  (assigning the values of the fluxes on the line at j = constant
c  and maximum error evaluation)
c
      do 25 i = 1,n
      oldphi(i,j) = phi(i,j)
      phi(i,j) = v(i)
c
      delphi = phi(i,j) - oldphi(i,j)
c     ratio = dabs (delphi) / ( phi(i,j) + 1.d-10 )
      ratio = dabs (delphi)
      if(ratio.gt.ratmax) ratmax = ratio
   25 continue
c
   30 continue
c
      if((ratmax.lt.1.d00).and.(iter.ne.1)) goto 45
c
      if(istvid.eq.1) write(*,101) iter,ratmax
      write(6,101) iter,ratmax
   40 continue
      write(6,100)
c
   45 return
  100 format (/,1x,'<<<<< Warning: Convergence Problems >>>>>')
  101 format (' LOR Internal Iteration n. ',i6,
     & ' Max. Error = ',e14.7)
      end
c----------------------------------------------------------------------c
c                                                                      c
c   Soluzione di Sistemi Tridiagonali con l'Algoritmo di Thomas        c
c   (Solution of systems with Tridiagonal matrix                       c
c    by the Thomas algorithm)                                          c
c                                                                      c
c----------------------------------------------------------------------c
      subroutine tdma (a,b,c,d,v,alef,bet,n,ld)
      implicit double precision (a-h,o-z)
      dimension a(ld),b(ld),c(ld),d(ld),v(ld),alef(ld),bet(ld)
      ub=1.d00/b(1)
      alef(1)=c(1)*ub
      bet(1)=d(1)*ub
      do 10 i=2,n
      l=i-1
      qz=b(i)-a(i)*alef(l)
      uqz=1.d00/qz
      alef(i)=c(i)*uqz
   10 bet(i)=(d(i)-a(i)*bet(l))*uqz
      nm1=n-1
      v(n)=bet(n)
      do 20 i=1,nm1
      ii=n-i
      l=ii+1
   20 v(ii)=bet(ii)-alef(ii)*v(l)
      return
      end
